/*****************************************************************************
 * Source code information
 * -----------------------
 * Original author    Ian Dickinson, HP Labs Bristol
 * Author email       Ian.Dickinson@hp.com
 * Package            Jena 2
 * Web                http://sourceforge.net/projects/jena/
 * Created            14-Mar-2003
 * Filename           $RCSfile: TestPath.java,v $
 * Revision           $Revision: 1.7 $
 * Release status     $State: Exp $
 *
 * Last modified on   $Date: 2004/12/06 13:50:30 $
 *               by   $Author: andy_seaborne $
 *
 * (c) Copyright 2002, 2003, 2004 Hewlett-Packard Development Company, LP
 * (see footer for full conditions)
 *****************************************************************************/

// Package
///////////////
package com.hp.hpl.jena.ontology.path.impl.test;


// Imports
///////////////
import junit.framework.*;

import com.hp.hpl.jena.ontology.*;
import com.hp.hpl.jena.ontology.path.*;
import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.vocabulary.*;


/**
 * <p>
 * Unit test cases for path expressions. 17th Nov 2003: Chris suppressed tests
 * testComposeUnitPath and testComposeComplexPath: they fail due to some
 * order sensitivity and Ian argues in faviour of ditching the whole Path tree anyway.
 * 
 * </p>
 *
 * @author Ian Dickinson, HP Labs
 *         (<a  href="mailto:Ian.Dickinson@hp.com" >email</a>)
 * @version CVS $Id: TestPath.java,v 1.7 2004/12/06 13:50:30 andy_seaborne Exp $
 */
public class TestPath 
    extends TestCase
{
    // Constants
    //////////////////////////////////

    protected static final String NS = "http://test/path#";
    
    
    // Static variables
    //////////////////////////////////

    protected static Object[][] s_stmts = new Object[][] {
        //subject   pred    object
        { "a",      "p",    "b"},
        { "a",      "q",    "c"},
        { "b",      "q",    "d0"},
        { "b",      "p",    "d1"},
        { "b",      "p",    "d2"},
        { "d1",      "q",    "e1"},
        { "d1",      "q",    "e2"},
        { "e",      "q",    "e" },
        { "e",      "p",    "e" },
        { "f0",     "q",    "f1" },
        { "f0",     "q",    "f2" },
        { "f1",     "q",    "f4" },
        { "f2",     "q",    "f4" },
        { "f2",     "q",    "f3" },
    };
    
    
    // Instance variables
    //////////////////////////////////
    
    protected OntModel m_model = ModelFactory.createOntologyModel( OWL.getURI() );
    
    protected Property m_p;
    protected Property m_q;
    protected Resource m_a;
    protected Resource m_b;
    protected Resource m_c;
    protected Resource m_e;
    
    
    // Constructors
    //////////////////////////////////

    public TestPath( String name ) {
        super( name );
    }
    
    
    // External signature methods
    //////////////////////////////////

    public void setUp() {
        // ensure the ont doc manager is in a consistent state
        OntDocumentManager.getInstance().reset( true );

        for (int i = 0; i < s_stmts.length;  i++) {
            Resource s = m_model.getResource( NS + ((String) s_stmts[i][0]) );
            Property p = m_model.getProperty( NS + ((String) s_stmts[i][1]) );
            Resource o = m_model.getResource( NS + ((String) s_stmts[i][2]) );
            
            m_model.add( s, p, o );
        }
        
        m_p = m_model.getProperty( NS + "p" );
        m_q = m_model.getProperty( NS + "q" );
        
        m_a = m_model.getResource( NS + "a" );
        m_b = m_model.getResource( NS + "b" );
        m_c = m_model.getResource( NS + "c" );
        m_e = m_model.getResource( NS + "e" );
    }
    
    public void tearDown() {
    }
    
    /** Test unit named paths */
    public void testNamedUnitPath() {
        PathExpr pe = PathFactory.unit( m_p );
        PathSet ps = pe.asPathSet( m_a );
        
        assertEquals( "Unit path set over p from a should have size 1", 1, ps.size() );
        assertTrue( "Path set should not be empty", !ps.isEmpty() );
        
        PathIterator i = ps.paths();
        assertTrue( "Should be at least one path over p from a", i.hasNext() );
        i.next();
        assertTrue( "Should be only one path over p from a", !i.hasNext() );
        
        assertEquals( "Value should be b", m_b, ps.getValue() );
        assertTrue( "should have value b", ps.hasValue( m_b ) );
        
        Path p = ps.paths().nextPath();
        assertEquals( "unit path should have length 1", 1, p.length() );
        
        assertTrue( "should contain 'a p b' ", m_model.contains( m_a, m_p, m_b ) );
        ps.removeAll();
        assertTrue( "should not contain 'a p b' ", !m_model.contains( m_a, m_p, m_b ) );
        
        // look up again - should be empty
        ps = pe.asPathSet( m_a );
        assertTrue( "Path set should now be empty", ps.isEmpty() );
        
        boolean ex = false;
        try {
            // should cause an exception as the path is empty
            ps.getValue();
        }
        catch (PathException e) {
            ex = true;
        }
        assertTrue( "Getting value from empty path should raise exception", ex );
    }
    
    
    /** Test unit any paths */
    public void testAnyUnitPath() {
        PathExpr pe = PathFactory.unit();
        PathSet ps = pe.asPathSet( m_a );
        
        assertEquals( "Unit path set over p from a should have size 2", 2, ps.size() );
        assertTrue( "Path set should not be empty", !ps.isEmpty() );
        
        PathIterator i = ps.paths();
        assertTrue( "Should be at least one path over p from a", i.hasNext() );
        i.next();
        assertTrue( "Should be at least two paths over p from a", i.hasNext() );
        i.next();
        assertTrue( "Should be only two paths over p from a", !i.hasNext() );
        
        assertTrue( "should have value b", ps.hasValue( m_b ) );
        assertTrue( "should have value c", ps.hasValue( m_c ) );
        
        Path p = ps.paths().nextPath();
        assertEquals( "unit path should have length 1", 1, p.length() );
        
        assertTrue( "should contain 'a p b' ", m_model.contains( m_a, m_p, m_b ) );
        assertTrue( "should contain 'a p b' ", m_model.contains( m_a, m_q, m_c ) );
        ps.removeAll();
        assertTrue( "should not contain 'a p b' ", !m_model.contains( m_a, m_p, m_b ) );
        assertTrue( "should not contain 'a p b' ", !m_model.contains( m_a, m_q, m_c ) );
        
        // look up again - should be empty
        ps = pe.asPathSet( m_a );
        assertTrue( "Path set should now be empty", ps.isEmpty() );
    }
    
    
    /** Test composed paths of unit paths */
    public void xxtestComposeUnitPath() {
        PathExpr pq = PathFactory.compose( m_p, m_q );
        PathSet pqSet = pq.asPathSet( m_a );
        
        assertTrue( "p o q from node a should not be empty", !pqSet.isEmpty() );
        assertEquals( "p o q from node a should have size 1", 1, pqSet.size() );
        
        // now try the same path from node b
        PathSet pqSetB = pq.asPathSet( m_b );
        
        assertTrue( "p o q from node b should not be empty", !pqSetB.isEmpty() );
        assertEquals( "p o q from node b should have size 2", 2, pqSetB.size() );
        
        Resource e2 = m_model.getResource( NS + "e2" );
        assertTrue( "p o q from node b should have e2 as a value", pqSetB.hasValue(e2) );
        
        // node e is reflexive on p and q
        PathSet pqSetE = pq.asPathSet( m_e );
        assertTrue( "p o q from node e should not be empty", !pqSetE.isEmpty() );
        assertEquals( "p o q from node e should have size 1", 1, pqSetE.size() );
        
    }
    
    
    /** Test multi-layer composed paths */
    public void xxtestComposeComplexPath() {
        PathExpr pp = PathFactory.compose( m_p, m_p );
        PathExpr ppq = PathFactory.compose( pp, m_q );
        
        assertEquals( "ppq size should be 2", 2, ppq.asPathSet( m_a ).size() );
        
        Resource e2 = m_model.getResource( NS + "e2" );
        assertTrue( "p o p o q should have e2 as value", ppq.asPathSet( m_a ).hasValue( e2 ));
    }
    
    
    /** Test closure paths on non-inferencing graph */
    public void testClosurePath() {
        PathExpr qstar = PathFactory.closure( m_q );
        
        assertEquals( "q* from a should have size 1", 1, qstar.asPathSet( m_a ).size() );
        
        Resource f0 = m_model.getResource( NS + "f0" );
        assertEquals( "q* from f0 should have size 5", 5, qstar.asPathSet( f0 ).size() );
        
        // with occurs check, path from e should be length one
        Resource e = m_model.getResource( NS + "e" );
        assertEquals( "q* from e should be size 1", 1, qstar.asPathSet( e ).size() );
        
        // without occurs check, is infinite
        int count = 0;
        PathIterator qqqe = PathFactory.closure( m_q, false ).evaluate( e );
        
        for (count = 0; count < 100; count++) {
            assertTrue( "Should be more on the path", qqqe.hasNext() );
            assertEquals( "Path length not correct", count + 1, qqqe.nextPath().length() );
        } 
        assertEquals( "Did not complete iteration", 100, count );
    }
    
    
    // Internal implementation methods
    //////////////////////////////////

    //==============================================================================
    // Inner class definitions
    //==============================================================================

}


/*
    (c) Copyright 2002, 2003, 2004 Hewlett-Packard Development Company, LP
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:

    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. The name of the author may not be used to endorse or promote products
       derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
